/**
 * Copyright (c) 2009 Carnegie Mellon University.
 * All rights reserved.
 *
 * Permission to use this software and its documentation for any purpose is
 * hereby granted, provided that the above copyright notice appear and that
 * both that copyright notice and this permission notice appear in supporting
 * documentation, and that the name of CMU not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.
 *
 * CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WSTRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL CMU BE
 * LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, RISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * GCC is free software; you can redistribute it and/or modify it under|
 * the terms of the GNU General Public License as published by the Free|
 * Software Foundation; either version 3, or (at your option) any later|
 * version.|
 * |
 * GCC is distributed in the hope that it will be useful, but WITHOUT ANY|
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or|
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License|
 * for more details.|
 * |
 * You should have received a copy of the GNU General Public License|
 * along with GCC; see the file COPYING3.  If not see|
 * <http://www.gnu.org/licenses/>.
 */

#include "air-internal.h"

struct whitelisted_file {
  /* The name of this file.  */
  char *file_name;
  /* The line numbers that are whitelisted in this file. These
  ** must be sorted.  */
  unsigned int *lines;
  /* The number of whitelisted lines.  */
  unsigned int num_lines;
};

typedef struct whitelisted_file white_file_t;

/* An array of the whitelisted file names. Theses file names
** must be sorted alphabetically.  */
white_file_t *whitelist = NULL;

/* The number of files that are partially whitelisted.  */
unsigned int whitelist_size = 0;

#define AIR_MAX_LINE_LENGTH 1024

/* Initialize the whitelist from the file given.  */
void
init_whitelist (const char *file_name)
{
  char buf[AIR_MAX_LINE_LENGTH];
  unsigned long int val;
  unsigned int i, j;
  white_file_t *ptr = NULL;
  FILE *whitelist_file = fopen (file_name, "r");
  if (!whitelist_file)
    internal_error ("opening whitelist file %s failed", file_name);

  if (fgets (buf, sizeof(buf), whitelist_file) == NULL)
    internal_error ("unable to read whitelist file count from %s", file_name);

  val = strtoul (buf, NULL, 10);
  if (val > UINT_MAX)
    internal_error ("whitelist file count out of range for %s", file_name);
  if (val == 0)
    return;
  whitelist_size = (unsigned int) val;

  whitelist = (white_file_t*) xmalloc (whitelist_size * sizeof (white_file_t));
  ptr = whitelist;

  for (i = 0; i < whitelist_size; i++, ptr++) {
    if (fgets (buf, sizeof(buf), whitelist_file) == NULL)
      internal_error ("expected file name in whitelist %s", file_name);
    
    val = strlen (buf);
    if (val > 0 && buf[val - 1] == '\n')
      buf[val - 1] = '\0';
    ptr->file_name = (char*) xmalloc (val + 1);
    strcpy (ptr->file_name, buf);

    if (fgets (buf, sizeof(buf), whitelist_file) == NULL)
      internal_error ("expected line count for %s in whitelist not found",
          ptr->file_name);
    val = strtoul (buf, NULL, 10);
    if (val > UINT_MAX)
      internal_error ("whitelist line count for %s out of range",
          ptr->file_name);
    ptr->num_lines = (unsigned int) val;

    ptr->lines = (unsigned int*) xmalloc (ptr->num_lines *
                                                (sizeof (unsigned int)));

    for (j = 0; j < ptr->num_lines; j++) {
      if (fgets (buf, sizeof(buf), whitelist_file) == NULL)
        internal_error ("expected line number in whitelist for %s",
            ptr->file_name);
      val = strtoul (buf, NULL, 10);
      if (val > UINT_MAX)
        internal_error ("whitelist line number for %s out of range",
            ptr->file_name);
      ptr->lines[j] = (unsigned int) val;
    }
  }
  if (fclose(whitelist_file) == EOF)
    internal_error ("unable to close whitelist %s", file_name);
}

/* Binary search for the index of the given file in the list
** of whitelisted files. Return the index if found, otherwise
** return whitelist_size.  */
static unsigned int
file_index (const char *file)
{
  unsigned int min = 0;
  unsigned int max = whitelist_size - 1;
  unsigned int mid;
  int cmp;
  
  while (min <= max) {
    mid = min + (max - min) / 2;
    cmp = strcmp (whitelist[mid].file_name, file);
    if (cmp < 0)
      min = mid + 1;
    else if (cmp > 0) {
      if (mid == 0)
        break;
      max = mid - 1;
    }
    else
      return mid;
  }
  return whitelist_size;
}

/* Binary search for the line in the given file. Return
** true if it is in the whitelist, false otherwise.  */
bool
whitelisted_location_p (const char *file, unsigned int line)
{
  if (!air_whitelist)
    return false;
  else {
    unsigned int i = file_index (file);
    unsigned int min = 0, mid, max;
    unsigned int *lines = NULL;

    if (i == whitelist_size)
      return false;
    lines = whitelist[i].lines;
    max = whitelist[i].num_lines - 1;

    while (min <= max) {
      mid = min + (max - min) / 2;
      if (lines[mid] < line)
        min = mid + 1;
      else if (lines[mid] > line) {
        if (mid == 0)
          break;
        max = mid - 1;
      }
      else
        return true;
    }
    return false;
  }
}

