/**
 * Copyright (c) 2009 Carnegie Mellon University.
 * All rights reserved.
 *
 * Permission to use this software and its documentation for any purpose is
 * hereby granted, provided that the above copyright notice appear and that
 * both that copyright notice and this permission notice appear in supporting
 * documentation, and that the name of CMU not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.
 *
 * CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WSTRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL CMU BE
 * LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, RISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * GCC is free software; you can redistribute it and/or modify it under|
 * the terms of the GNU General Public License as published by the Free|
 * Software Foundation; either version 3, or (at your option) any later|
 * version.|
 * |
 * GCC is distributed in the hope that it will be useful, but WITHOUT ANY|
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or|
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License|
 * for more details.|
 * |
 * You should have received a copy of the GNU General Public License|
 * along with GCC; see the file COPYING3.  If not see|
 * <http://www.gnu.org/licenses/>.
 */

#include "air-internal.h"
#include "air-rtl.h"
void *air_handle = NULL;

/* Change the sign of x according to the given sign and the
** current sign of x.  */
static void
set_sign(rtx x, unsigned int sign) {
  if (sign == RTX_WHITELIST || RTX_SIGN(x) == RTX_WHITELIST) {
    RTX_SIGN (x) = RTX_WHITELIST;
    return;
  }
  if (RTX_SIGN (x) == sign ||
      RTX_SIGN (x) == RTX_CONFLICT ||
      sign == RTX_NO_SIGN)
    return;
  else if (RTX_SIGN (x) == RTX_NO_SIGN)
    RTX_SIGN (x) = sign;
  else
    RTX_SIGN (x) = RTX_CONFLICT;
}

/* Merge the two signs of the operands to obtain a single sign for their
** result when an arithmetic operation is performed on them.  */
unsigned int
air_merge_signs (rtx op0, rtx op1) {
  if (flag_overflow_level) {
    unsigned int sign0 = op0 ? RTX_SIGN (op0) : RTX_NO_SIGN;
    unsigned int sign1 = op1 ? RTX_SIGN (op1) : RTX_NO_SIGN;
    if (sign0 == RTX_UNSIGNED || sign1 == RTX_UNSIGNED)
      return RTX_UNSIGNED;
    if (sign0 == RTX_SIGNED || sign1 == RTX_SIGNED)
      return RTX_SIGNED;
  }
  return RTX_NO_SIGN;
}

/* Use the boolean value unsignedp to set the sign of op0.  */
void
air_set_sign1 (unsigned int unsignedp, rtx op0)
{
  if (flag_overflow_level) {
    if (unsignedp)
      set_sign (op0, RTX_UNSIGNED);
    else
      set_sign (op0, RTX_SIGNED);
#ifdef AIR_DEBUG
    fprintf (stderr, "---------- In air_set_sign1 ----------\n");
    debug_rtx (op0);
#endif
  }
}

/* Use the boolean value unsignedp to set the sign of op0 and op1.  */
void
air_set_sign2 (unsigned int unsignedp, rtx op0, rtx op1)
{
  if (flag_overflow_level) {
    if (unsignedp) {
      set_sign (op0, RTX_UNSIGNED);
      set_sign (op1, RTX_UNSIGNED);
    }
    else {
      set_sign (op0, RTX_SIGNED);
      set_sign (op1, RTX_SIGNED);
    }
#ifdef AIR_DEBUG
    fprintf (stderr, "---------- In air_set_sign2 ----------\n");
    debug_rtx (op0);
    debug_rtx (op1);
#endif
  }
}

/* Use the boolean value unsignedp to set the sign of op0, op1,
** and op2.  */
void
air_set_sign3 (unsigned int unsignedp, rtx op0, rtx op1, rtx op2)
{
  if (flag_overflow_level) {
    if (unsignedp) {
      set_sign (op0, RTX_UNSIGNED);
      set_sign (op1, RTX_UNSIGNED);
      set_sign (op2, RTX_UNSIGNED);
    }
    else {
      set_sign (op0, RTX_SIGNED);
      set_sign (op1, RTX_SIGNED);
      set_sign (op2, RTX_SIGNED);
    }
#ifdef AIR_DEBUG
    fprintf (stderr, "---------- In air_set_sign3 ----------\n");
    debug_rtx (op0);
    debug_rtx (op1);
    debug_rtx (op2);
#endif
  }
}

/* Use the boolean value unsignedp to set the sign of op0, op1,
** op2, and op3.  */
void
air_set_sign4 (unsigned int unsignedp, rtx op0, rtx op1, rtx op2, rtx op3)
{
  if (flag_overflow_level) {
    if (unsignedp) {
      set_sign (op0, RTX_UNSIGNED);
      set_sign (op1, RTX_UNSIGNED);
      set_sign (op2, RTX_UNSIGNED);
      set_sign (op3, RTX_UNSIGNED);
    }
    else {
      set_sign (op0, RTX_SIGNED);
      set_sign (op1, RTX_SIGNED);
      set_sign (op2, RTX_SIGNED);
      set_sign (op3, RTX_SIGNED);
    }
#ifdef AIR_DEBUG
    fprintf (stderr, "---------- In air_set_sign4 ----------\n");
    debug_rtx (op0);
    debug_rtx (op1);
    debug_rtx (op2);
    debug_rtx (op3);
#endif
  }
}

/* Copy the sign from old_rtx to new_rtx.  */
void
air_copy_sign (rtx new_rtx, rtx old_rtx)
{
  if (flag_overflow_level) {
#ifdef AIR_DEBUG
    fprintf (stderr, "---------- In air_copy_sign ----------\n");
    debug_rtx (new_rtx);
    debug_rtx (old_rtx);
#endif
    set_sign (new_rtx, RTX_SIGN (old_rtx));
  }
}

/* Copy the whitelist status from old_rtx to new_rtx.  */
void
air_copy_whitelisted (rtx new_rtx, rtx old_rtx)
{
  if (flag_overflow_level) {
    new_rtx->whitelisted = old_rtx->whitelisted;
#ifdef AIR_DEBUG
    fprintf (stderr, "---------- In air_copy_whitelist ----------\n");
    debug_rtx (new_rtx);
    debug_rtx (old_rtx);
#endif
  }
}


/* Set the signs of op0 and op1 according to sign and binoptab, the
** type of operation being performed.  */
void
air_expand_binop (unsigned int unsignedp, optab binoptab, rtx op0, rtx op1)
{
  if (flag_overflow_level) {
    if ((binoptab == ashl_optab || binoptab == ssashl_optab) && !unsignedp)
      set_sign (op0, RTX_SIGNED);
    else if (binoptab == ssadd_optab || binoptab == usadd_optab ||
        binoptab == sssub_optab || binoptab == ussub_optab ||
        binoptab == ssmul_optab || binoptab == usmul_optab ||
        binoptab == add_optab || binoptab == sub_optab ||
        binoptab == smul_optab) {
      if (unsignedp) {
        set_sign (op0, RTX_UNSIGNED);
        set_sign (op1, RTX_UNSIGNED);
      }
      else {
        set_sign (op0, RTX_SIGNED);
        set_sign (op1, RTX_SIGNED);
      }
    }
#ifdef AIR_DEBUG
    fprintf (stderr, "---------- In expand_binop ----------\n");
    debug_rtx (op0);
    debug_rtx (op1);
#endif
  }
}

/* Like air_set_sign1, set the sign of op0. However, sign is an
** RTX_SIGN, not a boolean.  */
void
air_set_sign_direct (unsigned int sign, rtx op0)
{
  if (flag_overflow_level) {
    set_sign (op0, sign);
#ifdef AIR_DEBUG
    fprintf (stderr, "---------- In air_set_sign_direct ----------\n");
    debug_rtx (op0);
#endif
  }
}

void
air_whitelist_expr (tree exp)
{
  int index;
  int i;

  if (!exp || !EXPR_P(exp))
    return;
  index = TREE_OPERAND_LENGTH(exp);
  for (i = 0; i < index; i++) {
    tree subexp = TREE_OPERAND (exp, i);
    if (subexp && EXPR_P(subexp)) {
      switch (TREE_CODE (TREE_OPERAND(exp, i))) 
      {
         case PLUS_EXPR:
         case MULT_EXPR:
         case NEGATE_EXPR:
           OVERFLOW_WHITELIST (TREE_OPERAND(exp, i));
         default:
           break;
      }
    }
    air_whitelist_expr(TREE_OPERAND(exp, i));
  }
}

/* Whitelist op0 so that no operations performed on it are checked
** for overflow.  */
void
air_whitelist_rtx (rtx op0)
{
  if (flag_overflow_level && op0) {
    /*set_sign (op0, RTX_WHITELIST);*/
    op0->whitelisted = true;
#ifdef AIR_DEBUG
    fprintf (stderr, "---------- In air_whitelist_rtx ----------\n");
    debug_rtx (op0);
#endif
  }
}

/* Insert the runtime checks for truncation. NOTE - to_sign contains an
 * RTX_SIGN variable, while unsignedp (the sign of from) contains a boolean.
 */
void
air_emit_truncation_checks (unsigned int to_sign, enum machine_mode to_mode, rtx from,
    unsigned int unsignedp, enum machine_mode from_mode) {
  tree shift_amount = build_int_cst (NULL_TREE,
      to_sign == RTX_UNSIGNED ? 
      GET_MODE_BITSIZE (to_mode) :
      GET_MODE_BITSIZE (to_mode) - 1);
  rtx temp = expand_shift (RSHIFT_EXPR, from_mode, from, shift_amount, 0, 0);
  rtx no_trunc_label = gen_label_rtx ();
  rtx call = flag_check_overflow_nop ? gen_rtx_ASM_INPUT (VOIDmode, "nop") :
             gen_rtx_ASM_INPUT (VOIDmode, "call constraint_handler_wrapper_trunc");
  do_compare_rtx_and_jump (temp, CONST0_RTX(from_mode),
      EQ, 0, from_mode,
      NULL_RTX, NULL_RTX, no_trunc_label);
  if (!unsignedp && to_sign == RTX_SIGNED) {
    do_compare_rtx_and_jump (temp, gen_rtx_CONST_INT (from_mode, -1),
        EQ, 0, from_mode,
        NULL_RTX, NULL_RTX, no_trunc_label);
  }
  ASM_INPUT_SOURCE_LOCATION (call) = input_location;
  emit_call_insn (call);
  emit_label (no_trunc_label);
}

unsigned int air_get_tree_sign(tree op0, tree op1)
{
  unsigned int  sign = RTX_SIGNED;

  if (flag_overflow_level) {
    if (op1) {
      sign = TYPE_UNSIGNED(TREE_TYPE(op1));
    }
    if (op0 && sign == RTX_SIGNED) {
      sign = TYPE_UNSIGNED(TREE_TYPE(op0));
    }

#ifdef AIR_DEBUG
    fprintf (stderr, "---------- In air_get_tree_sign ----------\n");
    debug_tree (op0);
    debug_tree (op1);
#endif
      return sign;
    } else {
      return 1;
    }

}

