/**
 * Copyright (c) 2009 Carnegie Mellon University.
 * All rights reserved.
 *
 * Permission to use this software and its documentation for any purpose is
 * hereby granted, provided that the above copyright notice appear and that
 * both that copyright notice and this permission notice appear in supporting
 * documentation, and that the name of CMU not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.
 *
 * CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WSTRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL CMU BE
 * LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, RISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 * GCC is free software; you can redistribute it and/or modify it under|
 * the terms of the GNU General Public License as published by the Free|
 * Software Foundation; either version 3, or (at your option) any later|
 * version.|
 * |
 * GCC is distributed in the hope that it will be useful, but WITHOUT ANY|
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or|
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License|
 * for more details.|
 * |
 * You should have received a copy of the GNU General Public License|
 * along with GCC; see the file COPYING3.  If not see|
 * <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <stdio.h>

#include <assert.h>
#include <errno.h>
#include <pthread.h>
#include <string.h>

#include "air-errorcodes.h"

#ifndef true
#define true 1
#endif
#ifndef false
#define false 0
#endif

struct ptr_list {
  const void *ptr;
  struct ptr_list *next;
};
typedef struct ptr_list ptr_list_t;
pthread_mutex_t list_mutex = PTHREAD_MUTEX_INITIALIZER;

static ptr_list_t *head = NULL;
constraint_handler_t constraint_handler = notify_handler_s;

int shift_overflow_masks[] = {0x80000000, 0xc0000000, 0xe0000000};

/********* End prototypes ************/

/*
 * As defined in ISO/IEC TR 24731-1 6.6.1.1
 */
constraint_handler_t
set_constraint_handler_s (constraint_handler_t handler) {
	constraint_handler = handler;
	return constraint_handler;
}

/*
 * As defined in ISO/IEC TR 24731-1 6.6.1.2
 */
void
abort_handler_s (const char * __restrict_arr msg, void * __restrict_arr ptr,
                 errno_t error) {
	notify_handler_s(msg, ptr, error);
	abort();
}

void
notify_handler_s (const char * __restrict_arr msg, void * __restrict_arr ptr,
                  errno_t error) {
	if (ptr) {
		fprintf(stderr, "*** Runtime constraint violated:\n\t%s at address %p\n", msg, ptr);
	} else {
		fprintf(stderr, "*** Runtime constraint violated:\n\t%s at unknown address\n", msg);
	}
}

/*
 * As defined in ISO/IEC TR 24731-1 6.6.1.3
 */
void
ignore_handler_s (const char * __restrict_arr msg, void * __restrict_arr ptr,
                  errno_t error) {
	return;
}

static _Bool
list_contains(const void * ptr) {
	ptr_list_t *cur;
	for (cur = head; cur != NULL; cur = cur->next) {
		if (cur->ptr == ptr)
			return 1;
	}
	return 0;
}

static void
list_push(const void * ptr) {
	ptr_list_t *cur = (ptr_list_t *) malloc(sizeof(ptr_list_t));
	assert(cur != NULL);
	cur->ptr = ptr;
	cur->next = head;
	head = cur;
}

void
constraint_handler_wrapper_inner(void * ptr, opcode_sign_t sign, opcode_type_t type) {
	char err_str[256];
	err_str[0] = '\0';

	/***  LOCK  ***/
	pthread_mutex_lock(&list_mutex);
	if (list_contains(ptr))
		return;
	list_push(ptr);
	pthread_mutex_unlock(&list_mutex);
	/*** UNLOCK ***/

	switch (sign) {
		case OPCODE_SIGN_UNSIGNED:
			strcat(err_str, "Unsigned integer overflow in ");
			break;
		case OPCODE_SIGN_SIGNED:
			strcat(err_str, "Signed integer overflow in ");
			break;
		default:
			assert(type == OPCODE_TYPE_TRUNC);
			break;
	}

	switch (type) {
		case OPCODE_TYPE_ADD:
			strcat(err_str, "addition");
			break;
		case OPCODE_TYPE_SUB:
			strcat(err_str, "subtraction");
			break;
		case OPCODE_TYPE_MUL:
			strcat(err_str, "multiplication");
			break;
		case OPCODE_TYPE_SAL:
			strcat(err_str, "left-shift");
			break;
		case OPCODE_TYPE_INC:
			strcat(err_str, "increment");
			break;
		case OPCODE_TYPE_DEC:
			strcat(err_str, "decrement");
			break;
		case OPCODE_TYPE_NEG:
			strcat(err_str, "negation");
			break;
		case OPCODE_TYPE_TRUNC:
			strcat(err_str, "Truncation");
			break;
		default:
			assert(false);
			break;
	}

	constraint_handler(err_str, ptr ? ptr - 5 : ptr, 1);
}
